@@ -15,6 +15,8 @@ setOldClass(c("XMLInternalCommentNode", "XMLInternalNode"))
 setOldClass(c("XMLInternalElementNode", "XMLInternalNode"))
 setOldClass(c("XMLInternalTextNode", "XMLInternalNode"))
 
+setOldClass(c("XMLDTDNode", "XMLInternalNode"))
+
 setOldClass("XMLNamespace")
 
 setOldClass("XMLInternalDOM")
@@ -33,6 +35,7 @@ setAs("XMLInternalDOM", "character", function(from) saveXML(from))
 setAs("XMLInternalDocument", "XMLInternalNode",
        function(from) xmlRoot(from))
 
+
 setAs("XMLInternalNode", "XMLInternalDocument",
         function(from)
            .Call("R_getXMLNodeDocument", from)
@@ -92,7 +95,7 @@ xmlName.XMLInternalNode =
 function(node, full = FALSE)
 {
   ans = .Call("RS_XML_xmlNodeName", node)
-  if((is.logical(full) && full) || length(full)) {
+  if((is.logical(full) && full) || (!is.logical(full) && length(full))) {
     tmp = xmlNamespace(node)
     if(length(tmp) && length(names(tmp)) > 0 && names(tmp) != "")
        ans = paste(names(tmp), ans, sep = ":")
@@ -115,9 +118,9 @@ function(node, addNamespace = TRUE, ...)
 }
 
 xmlChildren.XMLInternalNode =
-function(x)
+function(x, addNames = TRUE)
 {
- .Call("RS_XML_xmlNodeChildrenReferences", x)
+ .Call("RS_XML_xmlNodeChildrenReferences", x, as.logical(addNames))
 }
 
 xmlSize.XMLInternalNode =
@@ -128,7 +131,10 @@ function(obj)
 function(x, i, j, ...)
 {
   kids = xmlChildren(x)
-  if(is.numeric(i))
+  if(length(kids) == 0)
+    return(NULL)
+  
+  if(is.numeric(i)) 
      kids[[i]]
   else {
      id = as.character(i)
@@ -149,8 +155,8 @@ function(x, i, j, ...)
      kids[i]
   else {
      id = as.character(i)
-     which = (id == sapply(kids, xmlName))
-     kids[which]
+     which = match(sapply(kids, xmlName), id)
+     kids[!is.na(which)]
   }
 }  
 
@@ -186,6 +192,7 @@ xmlParent =
 function(x)
  UseMethod("xmlParent")
 
+
 xmlParent.XMLInternalNode =
 function(x)
 {
@@ -237,17 +244,56 @@ newXMLDoc <-
 # Creates internal C-level libxml object for representing
 # an XML document/tree of nodes.
 #
-function(dtd, namespaces = NULL, addFinalizer = TRUE)
+function(dtd = NA, namespaces = NULL, addFinalizer = TRUE, name = character(), node = NULL) 
 {
   ans = .Call("R_newXMLDoc", dtd, namespaces)
 
   addDocFinalizer(ans, addFinalizer)
 
+  if(length(name))
+    docName(ans) = as.character(name)
+
+  if(length(node)) {
+#   if(missing(shallow)) {
+#     shallow = !is.null(as(node, "XMLInternalDocument"))
+#   }
+#    if(!deep) 
+#      .Call("R_insertXMLNode", ans, node, as.integer(NA), FALSE)
+    addChildren(ans, node)
+  }
+
   ans
 }
 
+XMLOptions = new.env()
+getOption =
+function(name, default = NULL, converter = NULL)
+{
+  if(!exists(name, XMLOptions, inherits = FALSE)) 
+    return(default)
+
+   ans = get(name, XMLOptions)
+
+   if(is.function(converter))
+     converter(ans)
+    else
+      ans
+}
+
+setOption =
+function(name, value)
+{
+   prev = getOption(name)
+   assign(name, value, XMLOptions)
+   prev
+}  
+
+
 newXMLNode <-
 #
+###XXX Note that there is another definition of this in dups.R
+ # Which is now elided.
+  
 # Create an internal C-level libxml node
 #
   #
@@ -258,13 +304,28 @@ newXMLNode <-
   # XXX Have to add something to force the namespace prefix into the node
   # when there is no corresponding definition for that prefix.
   
-function(name, ..., attrs = NULL, namespace = "", doc = NULL, .children = list(...))
+function(name, ..., attrs = NULL,
+         namespace = "", namespaceDefinitions = character(),
+         doc = NULL, .children = list(...), parent = NULL,
+         at = NA,
+         cdata = FALSE,
+         suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE) #  i.e. warn.
+        )
 {
     # make certain we have a character vector for the attributes.
  ids = names(attrs)
  attrs = as(attrs, "character")
  names(attrs) = ids
 
+   # Find any attributes that are actually namespace definitions.
+ i = grep("^xmlns", names(attrs))
+ if(length(i)) {
+   warning("Don't specify namespace definitions via 'attrs'; use namespaceDefinitions")
+   namespace = c(namespace, structure(attrs[i], names = gsub("^xmlns:", "", names(attrs)[i])))
+   attrs = attrs[ -i]
+ }
+ 
+
      # allow the caller to specify the node name as  ns_prefix:name
      # but we have to create it as name and the set the namespace.
  ns = nsPrefix = character()
@@ -274,68 +335,181 @@ function(name, ..., attrs = NULL, namespace = "", doc = NULL, .children = list(.
    name = name[2]
  }
 
+
+ if(missing(doc) && !missing(parent) &&
+     inherits(parent, "XMLInternalDocument")) {
+   doc = parent
+   parent = NULL
+ }
+ 
+ if(is.null(doc) && !is.null(parent)) 
+   doc = as(parent, "XMLInternalDocument")
+
+
    # create the node. Let's leave the namespace definitions and prefix till later.
 
   # xmlSetProp() routine in R_newXMLNode() handles namespaces on the attribute names,
   # even checking them.
- node <- .Call("R_newXMLNode", as.character(name), attrs, character(), doc)
+ node <- .Call("R_newXMLNode", as.character(name), attrs, character(), doc, namespaceDefinitions)
+
+ if(!is.null(parent))
+   addChildren(parent, node, at = at)
 
 
- nsDefs = list()
+if(TRUE) { # Create the name space definitions here rather than in C code.
+  nsDefs = lapply(seq(along  = namespaceDefinitions),
+                   function(i) {
+                                  
+                     newNamespace(node, namespaceDefinitions[[i]], names(namespaceDefinitions)[i], set = FALSE)
+                   })
+  names(nsDefs) = names(namespaceDefinitions)
+}
+else
+  nsDefs = xmlNamespaceDefinitions(node)
+
+   # Now that the namespaces are defined, we can define the attributes which _may_ use them.
+  addAttributes(node, .attrs = attrs)
+ 
+
  if(length(namespace) > 0) {
-    if(length(namespace) == 1 && length(names(namespace)) == 0) {
+      # a single element with no name so this is the prefix.
+    if(length(namespace) == 1 && length(names(namespace)) == 0 && length(grep("^(http|ftp):", namespace)) == 0) {
       if(namespace != "")
         ns = nsPrefix = namespace
     } else {
-      
+
+           # we have names and/or more than one element. So these are namespace definitions
       if(length(names(namespace)) == 0) 
         names(namespace) <- rep("", length(namespace))
 
+      if(length(namespace) > 1 && !is.na(match(namespace[1], names(namespace)[-1]))) {
+        if(length(ns)) 
+          warning("ignoring first element of namespace and using prefix from node name, ", ns)
+        else {
+          ns = namespace[1]
+          namespace = namespace[-1]
+        }
+      }
+        
+      
+
       if(length(namespace) > 1 && sum(names(namespace) == "") > 1)
         warning("more than one namespace to use as the default")
 
       nsDefs = lapply(seq(along = namespace),
                       function(i) {
                         prefix = names(namespace)[i]
-                        ns <- .Call("R_xmlNewNs", node, namespace[[i]], prefix)
+
+                        newNamespace(node, namespace[[i]], prefix)
                                 # Don't set the namespace. This is just a definition/declaration for
                                 # this node, but not necessarily the namespace to use for this node.
                                 # We set this below
-                        ns
                       })
+      names(nsDefs) = names(namespace)
     }
   }
 
 
    # Now handle the prefix for this node.
  if(length(ns)) {
-   i = match(ns, names(namespace))
+   i = match(ns, names(nsDefs))
    if(is.na(i)) {
-      ns <- .Call("R_xmlNewNs", node, "", ns)
+      if(!is.null(parent)) 
+        ns = findNamespaceDefinition(node, ns)
+      else {
+         if(is.character(suppressNamespaceWarning))
+            f = get(suppressNamespaceWarning, mode = "function")
+         else if(is.logical(suppressNamespaceWarning) && !suppressNamespaceWarning) {
+            f = warning 
+         } else
+            f = function(...) {}
+             
+         f("cannot find namespace definition because the node is not in the document and there are no local namespace definitions for this node")
+         attr(node, "xml:namespace") = ns
+         ns = NULL
+      }
+#      ns <- newNamespace(node, "", ns)
    } else
       ns <- nsDefs[[i]]
  } else  {
-   i = match("", names(namespace))
+   i = match("", names(nsDefs))
    ns = if(is.na(i)) NULL else nsDefs[[i]]
  }
 
-   # Here is where we set the namespace for this node.
- if(length(ns) && (inherits(ns, c("XMLNamespaceRef", "XMLNamespaceDeclaration")) || (is.character(ns) && ns != "")))
-     .Call("R_xmlSetNs", node, ns, FALSE)
 
  
- if(!is.null(.children)) {
-   for(i in .children) {
-     if(!is(i, "XMLInternalNode"))
-        i = as(i, "XMLInternalNode")
-     
-     .Call("R_insertXMLNode", i, node)
-   }
+   # Here is where we set the namespace for this node.
+ if(length(ns) && (inherits(ns, c("XMLNamespaceRef", "XMLNamespaceDeclaration")) || (is.character(ns) && ns != "")))
+     setXMLNamespace( node, ns) # should this be append = FALSE ?
+
+ if(!is.null(.children))  {
+   if(!is.list(.children))
+      .children = list(.children)
+   addChildren(node, kids = .children, cdata = cdata)
  }
 
  node
 }
 
+findNamespaceDefinition =
+  #
+  # Search up the node hierarchy looking for a namespace
+  # matching that prefix.
+  #
+function(node, namespace, error = TRUE)
+{
+  ptr = node
+  while(!is.null(ptr)) {
+     tmp = namespaceDeclarations(ptr, TRUE)
+     i = match(namespace, names(tmp))
+     if(!is.na(i))
+       return(tmp[[i]])
+     ptr = xmlParent(ptr)
+  }
+
+  if(error)
+    stop("no matching namespace definition for prefix ", namespace)
+
+  NULL
+}  
+
+setXMLNamespace =
+  #
+  # Set the specified namespace as the namespace for this
+  # node. 
+  # namespace can be a prefix in which case we find it in the
+  # definition in this node or its ancestors.
+  # Otherwise, we expect a name = value character vector giving the
+  # prefix and URI and we create a new namespace definition.
+  # Alternatively, if you already have the namespace reference object
+  # from earlier, you can pass that in.
+  # Then we set the namespace on the node. 
+function(node,  namespace, append = FALSE)
+{
+  if(is.character(namespace) && is.null(names(namespace))) 
+
+     namespace = findNamespaceDefinition(node, namespace)
+
+  else if(is.character(namespace))
+
+     namespace = newNamespace(node, namespace)
+
+  else if(!inherits(namespace, c("XMLNamespaceRef", "XMLNamespaceDeclaration")))
+    stop("Must provide a namespace definition, a prefix of existing namespace or a reference to a namespace definition")
+
+  .Call("R_xmlSetNs", node, namespace, FALSE) 
+}  
+
+newNamespace =
+  # Create a new namespace reference object.
+function(node, namespace, prefix = names(namespace), set = FALSE)
+{
+   ns <- .Call("R_xmlNewNs", node, namespace, prefix)
+   if(set)
+      setXMLNamespace(node, ns)
+   ns     
+}
+
 checkNodeNamespace =
   #
   # can only be checked after we know the parent node, 
@@ -369,47 +543,183 @@ function(node, prefix = xmlNamespace(ns))
 # 
 
 newXMLTextNode =
-function(text, doc = NULL)
-  .Call("R_newXMLTextNode", as.character(text), doc)      
+  #
+  #  cdata allows the caller to specify that the text be 
+  #  wrapped in a newXMLCDataNode
+function(text,  parent = NULL, doc = NULL, cdata = FALSE)
+{
+  if(cdata) 
+    return(newXMLCDataNode(text, parent, doc))
+
+  a = .Call("R_newXMLTextNode", as.character(text), doc)
+  
+  if(!is.null(parent))
+     addChildren(parent, a)
+
+  a
+}
 
 newXMLPINode <-
-function(name,  text, doc = NULL)
+function(name,  text,  parent = NULL, doc = NULL, at = NA)
 {
-  .Call("R_newXMLPINode", doc, as.character(name), as.character(text))
+  a = .Call("R_newXMLPINode", doc, as.character(name), as.character(text))
+  if(!is.null(parent))
+    addChildren(parent, a, at = at)
+  a  
 }
 
 newXMLCDataNode <-
-function(text, doc = NULL)
+function(text, parent = NULL, doc = NULL, at = NA)
 {
-  .Call("R_newXMLCDataNode", doc,  as.character(text))
+  a = .Call("R_newXMLCDataNode", doc,  as.character(text))
+  if(!is.null(parent))
+    addChildren(parent, a, at = at)
+  a  
 }
 
 newXMLCommentNode <-
-function(text, doc = NULL)  # doc is not used.XXX
+function(text, parent = NULL, doc = NULL, at = NA) 
 {
-  .Call("R_newXMLComment",  as.character(text))
+  a = .Call("R_xmlNewComment",  as.character(text), doc)
+  if(!is.null(parent))
+    addChildren(parent, a, at = at)
+  a  
 }
 
-
-
-
-addChildren =
-function(node, ..., kids = list(...))
+replaceNodes =
+function(oldNode, newNode, ...)
 {
-  for(i in kids) {
-     if(is.list(i)) {
-       for(k in i)
-         addChildren(node, k)
-     } else {
-       if(is.character(i))
-          i = newXMLTextNode(i)
-     .Call("R_insertXMLNode", i, node)
-     }
+  UseMethod("replaceNodes")
+}
+
+replaceNodes.XMLInternalNode =
+function(oldNode, newNode, ...)
+{
+  oldNode = as(oldNode, "XMLInternalNode")
+  newNode = as(newNode, "XMLInternalNode")
+  
+  .Call("RS_XML_replaceXMLNode", oldNode, newNode)
+}  
+
+"[[<-.XMLInternalNode" =
+function(x, i, j, ..., value)
+{
+   if(!is.list(value))
+     value = list(value)
+
+   addChildren(x, kids = value, at = i)
+  x
+}  
+
+
+addChildren.XMLInternalNode =
+addChildren.XMLInternalDocument =
+  #
+  # XXX need to expand/recycle the at if it is given as a scalar
+  # taking into account if the subsequent elements are lists, etc.
+  #
+  # Basically, if the caller specifies at as a scalar
+  # we expand this to be the sequence starting at that value
+  # and having length which is the total number of nodes
+  # in kids.  This is not just the length of kids but
+  # the number of nodes since some of the elements might be lists.
+  #
+function(node, ..., kids = list(...), at = NA, cdata = FALSE)
+{
+  kids = unlist(kids, recursive = FALSE)
+  
+  if(!is.na(at)) {
+
+       # if at is the name of a child node, find its index (first node with that name)
+    if(is.character(at)) 
+      at = match(at, names(node))
+
+    
+    if(length(at) == 1) 
+      at = seq(as.integer(at), length = sum(sapply(kids, function(x) if(is.list(x)) length(x) else 1)))
+    else  # pad with NAs
+      length(at) = length(kids)
+    
+    return(lapply(seq(along = kids),
+            function(j) {
+               i = kids[[j]]
+
+               if(is.character(i))
+                 i = newXMLTextNode(i, cdata = cdata)
+
+               if(!is(i, "XMLInternalNode"))
+                 i = as(i, "XMLInternalNode")
+
+               if(is.na(at[j]))
+                 .Call("R_insertXMLNode", i, node, -1L, FALSE)
+               else {
+                  after = at[j] > 0
+                  if(!after)
+                     at[j] = 1
+
+                  if(xmlSize(node) < at[j])
+                    .Call("R_insertXMLNode", i, node, as.integer(NA), FALSE)
+                  else
+                    .Call("RS_XML_xmlAddSiblingAt", node[[ at[j] ]], i, after) # if at = 0, then shove it in before the sibling.
+               }
+            }))
   }
 
+  for(j in seq(along = kids)) {
+      i = kids[[j]]
+      
+      if(is.list(i)) {  # can't happen now since we unlist()
+         for(k in i)
+            addChildren(node, k)
+      } else {
+
+        if(is.null(i))
+           next
+     
+        if(is.character(i))
+          i = newXMLTextNode(i, cdata = cdata)
+
+        if(!is(i, "XMLInternalNode"))
+           i = as(i, "XMLInternalNode")
+
+        .Call("R_insertXMLNode", i, node, at[j], FALSE)
+
+        ns = attr(i, "xml:namespace")
+        if(!is.null(ns)) {
+           nsdef = findNamespaceDefinition(node, ns)
+           if(!is.null(nsdef) && (inherits(nsdef, c("XMLNamespaceRef", "XMLNamespaceDeclaration")) || (is.character(nsdef) && nsdef != ""))) {
+             setXMLNamespace( i, nsdef)
+             attr(i, "xml:namespace") = NULL
+           }
+        }
+     }
+    }
+
   node
 }
 
+addSibling =
+function(node, ..., kids = list(...), after = NA)
+{
+  UseMethod("addSibling")
+}
+
+addSibling.XMLInternalNode =
+function(node, ..., kids = list(...), after = TRUE)  
+{
+   #XXX Why add as children?
+   if(FALSE && is.na(after))
+     addChildren(node, kids = kids, at = NA)
+   else {
+     lapply(kids,
+            function(x) {
+              .Call("RS_XML_xmlAddSiblingAt", node, x, as.logical(after))
+            })
+   }
+  
+}  
+
+
 
 removeNodes =
 function(node, free = rep(FALSE, length(node)))
@@ -419,7 +729,7 @@ function(node, free = rep(FALSE, length(node)))
       stop("removeNode only works on internal nodes at present")
 
     free = as.logical(free)
-    length(free) = length(node)
+    free = rep(free, length = length(node))
   } else if(!inherits(node, "XMLInternalNode"))
      stop("removeNode is intended only for internal nodes")
   else {            
@@ -494,11 +804,11 @@ function(node, ..., kids = list(...), free = FALSE)
 
 
 setGeneric("toHTML",
-            function(x) standardGeneric("toHTML"))
+            function(x, context = NULL) standardGeneric("toHTML"))
 
 
 setMethod('toHTML', 'vector',
-            function(x) {
+            function(x, context = NULL) {
               tb = newXMLNode("table")
               if(length(names(x)) > 0) 
                 addChildren(tb, newXMLNode("tr", .children = sapply(names(x), function(x) newXMLNode("th", x))))
@@ -509,7 +819,7 @@ setMethod('toHTML', 'vector',
             })
 
 setMethod('toHTML', 'matrix',
-            function(x) {
+            function(x, context = NULL) {
               tb = newXMLNode("table")
               if(length(colnames(x)) > 0) 
                 addChildren(tb, newXMLNode("tr", .children = sapply(names(x), function(x) newXMLNode("th", x))))
@@ -529,7 +839,23 @@ setMethod('toHTML', 'matrix',
 
 
 
+SpecialCallOperators =
+  c("+", "-", "*", "/", "%*%", "%in%", ":")
 
+#XXX Not necessarily working yet! See RXMLDoc
+setMethod('toHTML', 'call',
+            function(x, context) {
+                # handle special operators like +, -, :, ...
+              if(as.character(v[[1]]) %in% SpecialCallOperators) {
+
+              }
+
+              v = newXMLNode(x[[1]], "(")
+              for(i in v[-1]) 
+                 addChildren(v, toHTML( i , context))
+
+              v
+            })
 
 setAs("vector", "XMLInternalNode",
       function(from) {
@@ -545,13 +871,19 @@ function(doc, file=NULL, compression=0, indent=TRUE, prefix = '<?xml version="1.
 }
 
 saveXML.XMLInternalNode <-
-function(doc, file=NULL, compression=0, indent=TRUE, prefix = '<?xml version="1.0"?>\n',
+function(doc, file = NULL, compression = 0, indent = TRUE, prefix = '<?xml version="1.0"?>\n',
          doctype = NULL, encoding = "")  
 {
   if(encoding == "")
     encoding = character()
-  
-  .Call("RS_XML_printXMLNode", doc, as.integer(0), as.integer(indent), as.logical(indent), as.character(encoding))
+
+  ans = .Call("RS_XML_printXMLNode", doc, as.integer(0), as.integer(indent), as.logical(indent), as.character(encoding))
+
+  if(length(file)) {
+    cat(ans, file = file)
+    file
+  } else
+    ans
 }
 
 setAs("XMLInternalNode", "character",
@@ -575,8 +907,16 @@ function(doc, file = NULL, compression = 0, indent = TRUE,
      prefix = c(doctype@name, doctype@public, doctype@system)
   }
 
-  .Call("R_saveXMLDOM", doc, file, as.integer(compression), as.logical(indent),
-                         as.character(prefix), as.character(encoding))
+  if(length(file))
+    file = path.expand(file)
+  
+  ans = .Call("R_saveXMLDOM", doc, file, as.integer(compression), as.logical(indent),
+                               as.character(prefix), as.character(encoding))
+
+  if(length(file))
+    file
+  else
+    ans
 }
 
 saveXML.XMLInternalDOM <-
@@ -602,7 +942,6 @@ saveXML.sink =
 function(doc, file = NULL, compression = 0, indent = TRUE, prefix = '<?xml version="1.0"?>\n',
          doctype = NULL, encoding = "")
 {
-
   if(inherits(file, c("character", "connection"))) {
     sink(file)
     on.exit(sink())
@@ -624,31 +963,54 @@ saveXML.XMLNode = saveXML.sink
 saveXML.XMLFlatTree = saveXML.sink
 
 
-setGeneric("addAttributes", function(node, ..., .attrs = NULL) standardGeneric("addAttributes"))
+setGeneric("addAttributes",
+           function(node, ..., .attrs = NULL, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE))
+             standardGeneric("addAttributes"))
 
 setMethod("addAttributes", "XMLInternalElementNode",
-function(node, ..., .attrs = NULL)
+function(node, ..., .attrs = NULL, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE) )
 {
-   if(missing(.attrs)) {
+   if(missing(.attrs)) 
      .attrs = list(...)
-   }
+
    .attrs = structure(as.character(.attrs), names = names(.attrs))
 
+   if(length(.attrs) == 0)
+      return(node)
+   
    if(is.null(names(.attrs)) || any(names(.attrs) == ""))
      stop("all node attributes must have a name")
 
+
+   if(is.character(suppressNamespaceWarning) || !suppressNamespaceWarning) {
+      nsDefs = xmlNamespaceDefinitions(node)
+      ns = sapply(strsplit(names(.attrs), ":"),
+                   function(x)  if(length(x) > 1) x[1] else NA)
+      i = which(!is.na(ns))
+      m = match(ns[i], names(nsDefs))
+      if(any(is.na(m))) {
+         f = if(is.character(suppressNamespaceWarning))
+                get(suppressNamespaceWarning, mode = "function")
+             else
+                warning
+
+         f(paste("missing namespace definitions for prefix(es)", paste(ns[i][is.na(m)])))
+      }
+   }
+
+   
    .Call("RS_XML_addNodeAttributes", node, .attrs)
    node
 })
 
 if(!isGeneric("xmlAttrs<-"))
- setGeneric("xmlAttrs<-", function(node, value)
+ setGeneric("xmlAttrs<-", function(node, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE), value)
                           standardGeneric("xmlAttrs<-"))
 
 tmp =
-function(node, value)
+function(node, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE), value)
 {
-   addAttributes(node, .attrs = value)
+   addAttributes(node, .attrs = value, suppressNamespaceWarning = suppressNamespaceWarning)
    node
 }
 
@@ -657,10 +1019,10 @@ setMethod("xmlAttrs<-", "XMLNode", tmp)
 
 
 setMethod("addAttributes", "XMLNode",
-           function(node, ..., .attrs = NULL) {
-             if(missing(.attrs)) {
+           function(node, ..., .attrs = NULL, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE) ) {
+             if(missing(.attrs)) 
                .attrs = list(...)
-             }
+
              .attrs = structure(as.character(.attrs), names = names(.attrs))
 
              if(is.null(names(.attrs)) || any(names(.attrs) == ""))
@@ -712,7 +1074,7 @@ function(node, ..., .attrs = NULL, .namespace = FALSE,
   
    if(.all) {
      if(length(list(...)) || length(.attrs))
-       stop(".all specified as TRUE and individual values specified via .../.attrs")
+         stop(".all specified as TRUE and individual values specified via .../.attrs")
      .attrs = names(xmlAttrs(node))
    }
    
@@ -812,6 +1174,8 @@ function(x, value)
    x
 }  
 
+
+
 newXMLNamespaces =
   # allow for multiple namespaces
   # and also allow for "r:value"
@@ -822,7 +1186,7 @@ function(node, ..., .values = list(...))
 {
   ids = names(.values)
   ans = lapply(ids, function(id)
-                      .Call("R_xmlNewNs", id, as.character(.values[[id]])))
+                      newNamespace(node, id, as.character(.values[[id]])))
 
   names(ans) = ids
   ans
